from . import __version__
VERSION_STRING = f'territory CLI {__version__}'


from argparse import ArgumentParser
from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryDirectory
import logging
import tarfile

import requests
import tqdm

from .api_client import DEFAULT_UPLOAD_TOKEN_PATH, auth, create_build_request
from .git import find_repo_root, list_repo_files, get_sha, get_commit_message, get_branch
from . import c, go
from .files import add_path_to_archive


def main(argv=None):
    args = parser.parse_args(args=argv)

    if args.L:
        logging.basicConfig(level=logging.DEBUG)
        requests_log = logging.getLogger("requests.packages.urllib3")
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True

    print(VERSION_STRING)
    cwd = (args.C or Path.cwd()).resolve()
    args.func(args, cwd)


@dataclass
class Package:
    work_dir: Path
    temp_dir: Path
    repo_root: Path
    captured_files: set[Path]
    index_system: bool
    upload_token: str | None


def upload(args, cwd):
    upload_token = None
    if not args.tarball_only:
        upload_token = auth(args)

    repo_root = find_repo_root(cwd)
    print('repository root directory:', repo_root)

    if args.l == 'go':
        lang = go.Lang()
    else:
        lang = c.Lang()

    with TemporaryDirectory() as td:
        td = Path(td)
        captured_files: set[Path] = set()

        tfl = Path(td, 'TERRITORY_FILE_LISTING')
        repo_files = list_repo_files(cwd)
        tfl.write_text(repo_files)
        captured_files.update(Path(cwd, p) for p in repo_files.split('\n'))

        package = Package(
            work_dir=cwd,
            temp_dir=td,
            repo_root=repo_root,
            captured_files=captured_files,
            index_system=args.system,
            upload_token=upload_token,
        )
        lang.prepare_package(package)

        if args.tarball_only:
            tarball_in = repo_root
        else:
            tarball_in = td
        tarball_path = Path(tarball_in, 'territory_upload.tar.gz')
        added = set()
        with tarfile.open(tarball_path, 'w:gz') as output:
            output.add(tfl, arcname=repo_root / 'TERRITORY_FILE_LISTING')
            lang.add_to_tar_file(package, output)
            for path in tqdm.tqdm(package.captured_files, 'compressing'):
                if not path.exists():
                    print('missing file:', path)
                    continue
                add_path_to_archive(added, output, path)

        if args.tarball_only:
            print('created', tarball_path)
            return

        print('collecting commit info')
        branch = str(get_branch(repo_root))
        meta = {
            'commit': get_sha(repo_root),
            'commit_message': get_commit_message(repo_root),
            'repo_root': str(repo_root),
            'index_system': args.system,
        }
        lang.add_to_meta(meta)

        print('registering build request')
        blob_size = tarball_path.stat().st_size
        intent = create_build_request(
            upload_token, args.repo_id, branch, meta, blob_size)

        print('uploading')
        with tarball_path.open('rb') as f:
            resp = requests.put(intent['url'], data=f, headers=intent['extensionHeaders'])
        resp.raise_for_status()

    if args.repo_id:
        print(f'Indexing will begin shortly. You can track build status at <https://app.territory.dev/repos/{args.repo_id}/jobs>.')


parser = ArgumentParser()
parser.add_argument('-C', type=Path, help='execute in a directory', metavar='PARSEDIR')
parser.add_argument('-L', action='store_true', help='enable debug logging')
parser.add_argument('--version', action='version', version=VERSION_STRING)

subparsers = parser.add_subparsers(required=True)
sp = subparsers.add_parser('upload')
sp.set_defaults(func=upload)
sp.add_argument('-l', choices=['c', 'c++', 'go'], help='Language to parse')
sp.add_argument(
    '--upload-token-path',
    type=Path,
    default=DEFAULT_UPLOAD_TOKEN_PATH)
sp.add_argument(
    '--system',
    action='store_true',
    help='collect system-wide dependencies')
repo_or_tarball = sp.add_mutually_exclusive_group(required=True)
repo_or_tarball.add_argument('--repo-id')
repo_or_tarball.add_argument(
    '--tarball-only',
    action='store_true',
    help='do not upload, create territory_upload.tar.gz output only')



from territory.c import collect_details, read_compile_commands, remove_arg, parse_vee
from territory_testlib import init_repo


def test_collect_details(tmp_path):
    init_repo(tmp_path / 'repo')
    cc_data = read_compile_commands(tmp_path / 'repo/compile_commands.json')
    td = tmp_path / 't'
    td.mkdir()
    paths = collect_details(td, tmp_path / 'repo', cc_data)
    assert '-target' in cc_data[0]['arguments']
    assert '-target' in cc_data[1]['arguments']
    assert tmp_path / 'repo/mod1.c' in paths
    assert tmp_path / 'repo/shared.h' in paths
    assert tmp_path / 'repo/dir/mod2.c' in paths


def test_collect_details_cc_error(tmp_path):
    init_repo(tmp_path / 'repo')
    (tmp_path / 'repo/mod1.c').write_text('BAD SYNTAX')
    (tmp_path / 'repo/dir/mod2.c').write_text('BAD SYNTAX')

    cc_data = read_compile_commands(tmp_path / 'repo/compile_commands.json')
    td = tmp_path / 't'
    td.mkdir()
    paths = collect_details(td, tmp_path / 'repo', cc_data)
    assert '-target' in cc_data[0]['arguments']
    assert '-target' in cc_data[1]['arguments']
    assert tmp_path / 'repo/mod1.c' in paths
    assert tmp_path / 'repo/dir/mod2.c' in paths
    assert tmp_path / 'repo/shared.h' not in paths


def test_remove_arg():
    assert remove_arg(['cc', '-c', '-o', '/foo', 'bar'], '-o', 2, prefix=True) == ['cc', '-c', 'bar']
    assert remove_arg(['cc', '-c', '-o/foo', 'bar'], '-o', 2, prefix=True) == ['cc', '-c', 'bar']
    assert remove_arg(['cc', '-c', '-o', '/foo', 'bar'], '-o', 2) == ['cc', '-c', 'bar']
    assert remove_arg(['cc', '-c', '-o/foo', 'bar'], '-o', 2) == ['cc', '-c', '-o/foo', 'bar']
    assert remove_arg(['cc', '-c', '-o', '/foo', 'bar'], '-c', 1) == ['cc', '-o', '/foo', 'bar']
    assert remove_arg(['cc', '-Ifoo', '-Ibar', '-Ibaz', '-iLocal', 'f.cc'], '-I', 2, prefix=True) == ['cc', '-iLocal', 'f.cc']


def test_read_vee():
    cc_output = '''Apple clang version 15.0.0 (clang-1500.3.9.4)
Target: arm64-apple-darwin23.6.0
Thread model: posix
InstalledDir: /Applications/Xcode.app/Contents/Developer/Toolchains/XcodeDefault.xctoolchain/usr/bin
 "/Applications/Xcode.app/Contents/Developer/Toolchains/XcodeDefault.xctoolchain/usr/bin/clang" -cc1 -triple arm64-apple-macosx14.0.0 -Wundef-prefix=TARGET_OS_ -Wdeprecated-objc-isa-usage -Werror=deprecated-objc-isa-usage -Werror=implicit-function-declaration -emit-obj -mrelax-all --mrelax-relocations -disable-free -clear-ast-before-backend -disable-llvm-verifier -discard-value-names -main-file-name test.c -mrelocation-model pic -pic-level 2 -mframe-pointer=non-leaf -fno-strict-return -ffp-contract=on -fno-rounding-math -funwind-tables=1 -fobjc-msgsend-selector-stubs -target-sdk-version=14.5 -fvisibility-inlines-hidden-static-local-var -target-cpu apple-m1 -target-feature +v8.5a -target-feature +crc -target-feature +lse -target-feature +rdm -target-feature +crypto -target-feature +dotprod -target-feature +fp-armv8 -target-feature +neon -target-feature +fp16fml -target-feature +ras -target-feature +rcpc -target-feature +zcm -target-feature +zcz -target-feature +fullfp16 -target-feature +sm4 -target-feature +sha3 -target-feature +sha2 -target-feature +aes -target-abi darwinpcs -debugger-tuning=lldb -target-linker-version 1053.12 -v -fcoverage-compilation-dir=/Users/pawel/territory -resource-dir /Applications/Xcode.app/Contents/Developer/Toolchains/XcodeDefault.xctoolchain/usr/lib/clang/15.0.0 -isysroot /Applications/Xcode.app/Contents/Developer/Platforms/MacOSX.platform/Developer/SDKs/MacOSX.sdk -I/usr/local/include -internal-isystem /Applications/Xcode.app/Contents/Developer/Platforms/MacOSX.platform/Developer/SDKs/MacOSX.sdk/usr/local/include -internal-isystem /Applications/Xcode.app/Contents/Developer/Toolchains/XcodeDefault.xctoolchain/usr/lib/clang/15.0.0/include -internal-externc-isystem /Applications/Xcode.app/Contents/Developer/Platforms/MacOSX.platform/Developer/SDKs/MacOSX.sdk/usr/include -internal-externc-isystem /Applications/Xcode.app/Contents/Developer/Toolchains/XcodeDefault.xctoolchain/usr/include -Wno-reorder-init-list -Wno-implicit-int-float-conversion -Wno-c99-designator -Wno-final-dtor-non-final-class -Wno-extra-semi-stmt -Wno-misleading-indentation -Wno-quoted-include-in-framework-header -Wno-implicit-fallthrough -Wno-enum-enum-conversion -Wno-enum-float-conversion -Wno-elaborated-enum-base -Wno-reserved-identifier -Wno-gnu-folding-constant -fdebug-compilation-dir=/Users/pawel/territory -ferror-limit 19 -stack-protector 1 -fstack-check -mdarwin-stkchk-strong-link -fblocks -fencode-extended-block-signature -fregister-global-dtors-with-atexit -fgnuc-version=4.2.1 -fmax-type-align=16 -fcommon -fcolor-diagnostics -clang-vendor-feature=+disableNonDependentMemberExprInCurrentInstantiation -fno-odr-hash-protocols -clang-vendor-feature=+enableAggressiveVLAFolding -clang-vendor-feature=+revert09abecef7bbf -clang-vendor-feature=+thisNoAlignAttr -clang-vendor-feature=+thisNoNullAttr -mllvm -disable-aligned-alloc-awareness=1 -D__GCC_HAVE_DWARF2_CFI_ASM=1 -o /tmp/test.o -x c /tmp/test.c
clang -cc1 version 15.0.0 (clang-1500.3.9.4) default target arm64-apple-darwin23.6.0
ignoring nonexistent directory "/Applications/Xcode.app/Contents/Developer/Platforms/MacOSX.platform/Developer/SDKs/MacOSX.sdk/usr/local/include"
ignoring nonexistent directory "/Applications/Xcode.app/Contents/Developer/Platforms/MacOSX.platform/Developer/SDKs/MacOSX.sdk/Library/Frameworks"
#include "..." search starts here:
#include <...> search starts here:
 /tmp/dir with space
 /usr/local/include
 /Applications/Xcode.app/Contents/Developer/Toolchains/XcodeDefault.xctoolchain/usr/lib/clang/15.0.0/include
 /Applications/Xcode.app/Contents/Developer/Platforms/MacOSX.platform/Developer/SDKs/MacOSX.sdk/usr/include
 /Applications/Xcode.app/Contents/Developer/Toolchains/XcodeDefault.xctoolchain/usr/include
 /Applications/Xcode.app/Contents/Developer/Platforms/MacOSX.platform/Developer/SDKs/MacOSX.sdk/System/Library/Frameworks (framework directory)
End of search list.
'''
    vee = parse_vee(cc_output)
    assert vee.target == 'arm64-apple-darwin23.6.0'
    assert vee.angle_bracket_include_paths == [
        '/tmp/dir with space',
        '/usr/local/include',
        '/Applications/Xcode.app/Contents/Developer/Toolchains/XcodeDefault.xctoolchain/usr/lib/clang/15.0.0/include',
        '/Applications/Xcode.app/Contents/Developer/Platforms/MacOSX.platform/Developer/SDKs/MacOSX.sdk/usr/include',
        '/Applications/Xcode.app/Contents/Developer/Toolchains/XcodeDefault.xctoolchain/usr/include',
        '/Applications/Xcode.app/Contents/Developer/Platforms/MacOSX.platform/Developer/SDKs/MacOSX.sdk/System/Library/Frameworks',
    ]



from hashlib import blake2b
from dataclasses import dataclass
from multiprocessing import cpu_count
from multiprocessing.pool import Pool
from pathlib import Path
from subprocess import PIPE, run
import json
import re
import shlex

import tqdm

from .files import find_in_ancestors


class Lang:
    def prepare_package(self, package):
        self.compile_commands_dir = find_compile_commands_dir(package.work_dir)
        print('compilation database:', self.compile_commands_dir / 'compile_commands.json')

        self.cc_path = self.compile_commands_dir / 'compile_commands.json'
        cc_data = read_compile_commands(self.cc_path)
        cc_files = collect_details(package.temp_dir, self.compile_commands_dir, cc_data)
        package.captured_files.update(cc_files)
        self.gen_ccs_path = Path(package.temp_dir, 'compile_commands.json')
        with self.gen_ccs_path.open('w') as f:
            json.dump(cc_data, f, indent=4)

    def add_to_tar_file(self, package, output):
        output.add(self.gen_ccs_path, arcname=self.cc_path)

    def add_to_meta(self, meta):
        meta['compile_commands_dir'] = str(self.compile_commands_dir)
        meta['lang'] = 'c'


def find_compile_commands_dir(p):
    try:
        return find_in_ancestors(p, lambda p: (p / 'compile_commands.json').exists())
    except FileNotFoundError:
        raise SystemExit('no compile_commands.json found')


def read_compile_commands(cc_path):
    with cc_path.open('r') as f:
        cc_data = json.load(f)

    for cc in cc_data:
        try:
            cmd_str = cc.pop('command')
        except KeyError:
            continue
        cc['arguments'] = shlex.split(cmd_str)

    return cc_data


def collect_details(tmp_dir, cc_dir, cc_data):
    with \
            Pool(cpu_count() * 2) as pool, \
            tqdm.tqdm(total=len(cc_data), desc='collecting compilation details') as progr:
        dep_paths = set()
        def _cb(details):
            idx, paths, arguments = details
            dep_paths.update(paths)
            cc_data[idx]['arguments'] = arguments
            progr.update(1)
        def _ecb(e):
            print('error:', e)
            progr.update(1)
        for i, cmd in enumerate(cc_data):
            dir_ = cmd.get('directory') or cc_dir
            p = Path(dir_, cmd['file'])
            dep_paths.add(p)
            pool.apply_async(
                _query_details,
                (i, cc_dir, tmp_dir, cmd),
                {},
                callback=_cb,
                error_callback=_ecb)
        pool.close()
        pool.join()

    return dep_paths


def remove_arg(arguments, key, count, prefix=False):
    res = []
    fi = 0
    while fi < len(arguments):
        arg = arguments[fi]

        if arg == key:
            fi += count

        elif  (prefix and arg.startswith(key)):
            fi += 1

        else:
            res.append(arg)
            fi += 1

    return res


@dataclass
class Vee:
    target: str | None = None
    angle_bracket_include_paths: list[str] | None = None


def parse_vee(text) -> Vee:
    vee = Vee()
    m = re.search('^Target: (.+)$', text, re.MULTILINE)
    if m:
        vee.target = m.group(1)

    m = re.search(r'#include <\.\.\.> search starts here:\n((^ .*\n)*)', text, re.MULTILINE)
    if m:
        vee.angle_bracket_include_paths = re.findall(r'^ (.*?)(?: \(framework directory|headermap\)\n|\n)', m.group(1), re.MULTILINE)
    return vee


def _query_details(index: int, cc_dir: Path, tmp_dir: Path, compilation_command):
    q_arguments = compilation_command['arguments'][:]

    q_arguments = remove_arg(q_arguments, '-c', 1)
    q_arguments = remove_arg(q_arguments, '-M', 1)
    q_arguments = remove_arg(q_arguments, '-MD', 1)
    q_arguments = remove_arg(q_arguments, '-MM', 1)
    q_arguments = remove_arg(q_arguments, '-MMD', 1)
    q_arguments = remove_arg(q_arguments, '-o', 2, prefix=True)
    q_arguments = remove_arg(q_arguments, '-MF', 2, prefix=True)

    deps_dir = tmp_dir / 'deps'
    deps_dir.mkdir(parents=True, exist_ok=True)
    deps_file = deps_dir / (blake2b(compilation_command['file'].encode()).hexdigest() + '.d')

    q_arguments = [q_arguments[0], '-E', '-MD', '-MF' + str(deps_file), *q_arguments[1:], '-v', '-o', '/dev/null', '-Wno-error']
    completion = run(q_arguments, cwd=compilation_command.get('directory') or cc_dir, stderr=PIPE, text=True)

    arguments = compilation_command['arguments'][:]
    vd = parse_vee(completion.stderr)
    if vd.target is not None:
        arguments[1:1] = ['-target', vd.target]

    if vd.angle_bracket_include_paths:
        arguments = remove_arg(arguments, '-I', 2, prefix=True)
        arguments = remove_arg(arguments, '--include-directory', 2)
        arguments = remove_arg(arguments, '--include-directory=', 1, prefix=True)
        arguments = remove_arg(arguments, '-cxx-isystem', 2, prefix=True)
        arguments = remove_arg(arguments, '-ibuiltininc', 1)
        arguments = remove_arg(arguments, '-iframework', 2, prefix=True)
        arguments = remove_arg(arguments, '-iframeworkwithsysroot', 2, prefix=True)
        arguments = remove_arg(arguments, '--stdlib++-isystem', 2, prefix=True)
        arguments = remove_arg(arguments, '-isystem', 2, prefix=True)
        arguments = remove_arg(arguments, '-M', 1)
        arguments = remove_arg(arguments, '-MD', 1)
        arguments = remove_arg(arguments, '-MM', 1)
        arguments = remove_arg(arguments, '-MMD', 1)
        arguments = remove_arg(arguments, '-MF', 2, prefix=True)

        incs = [f'-I{dir}' for dir in vd.angle_bracket_include_paths]
        arguments[1:1] = ['-nostdinc', *incs]

    if deps_file.exists():
        deps_text = deps_file.read_text()
        deps_file.unlink()
        try:
            _target, deps = deps_text.split(':', 1)
        except Exception as e:
            print('failed to read dependencies:', deps_text, e)
            return index, set(), arguments
        lines = [l.rstrip('\\') for l in deps.splitlines()]
        files = shlex.split(' '.join(lines))

        dir_ = compilation_command.get('directory') or cc_dir

        return index, {Path(dir_, f) for f in files}, arguments
    else:
        print('no dependencies recorded for', compilation_command['file'])

    if completion.returncode != 0:
        print(completion.stderr)
    return index, set(), arguments



from pathlib import Path
import tarfile


def find_in_ancestors(p: Path, f, highest=False):
    found = None
    for p in [p, *p.parents]:
        if f(p):
            if highest:
                found = p
            else:
                return p.resolve()
    if found:
        return found
    raise FileNotFoundError()


def add_path_to_archive(added: set[Path], archive: tarfile.TarFile, path: Path):
    '''Adds a file to archive, ensuring symlinks are preserved and paths normalized'''
    stk = list(path.parts)
    i = 1
    while i <= len(stk):
        p = Path(*stk[:i])
        if stk[i-1] == '..':
            stk[i-2 : i] = []
            i -= 2
        elif p.is_symlink():
            if p not in added:
                added.add(p)
                archive.add(p)
            lp = list(p.readlink().parts)
            stk[i-1:i] = lp
            i -= 1
        elif p not in added:
            added.add(p)
            archive.add(p, recursive=False)
        i += 1
